/* Example 2 -- a minimal RTP Listener.
 * Listen to the specified address, and print all events that occur.
 * 
 * Copyright 1997, 1998 Lucent Technologies, all rights reserved.
 *
 * BUGS: This assumes there is only one outstanding event scheduled
 * by RTPSchedule() at any one time.  This is currently true of the
 * library, but not guaranteed by the API.
 */
   
#ifndef RTPRECEIVE_H
#define RTPRECEIVE_H
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>    /* pthread */

#ifdef __unix
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <sys/time.h>  /* struct timeval, gettimeofday() */
#include <sys/param.h> /* Might declare MAXHOSTNAMELEN */
#include <netdb.h>     /* Might declare MAXHOSTNAMELEN */
#include <unistd.h>    /* Might declare gethostname() */
#endif

#ifdef WIN32
#include <winsock2.h>

#endif

#include "rtp_api.h"
#include "rtp_highlevel.h"

/* For windows, this is linked in via the rtp library */
#ifdef WIN32
extern void gettimeofday(struct timeval *, void *);
#endif

/* Convert a struct timeval to a double */
#define tv2dbl(tv) ((tv).tv_sec + (tv).tv_usec / 1000000.0)

#define PERIODTIME 1000
/* Convert a double to a struct timeval */
 struct timeval dbl2tv(double d);

/* Prints out reception reports */

void PrintReporting(context cid);

struct evt_queue_elt {
  context cid;
  rtp_opaque_t event_opaque;
  double event_time;
  struct evt_queue_elt *next;
};

struct png_queue{
  u_int32 ts;
  char *packet;         /* payload portion of RTP packet */
  long len;        /* The length of the payload */
  struct png_queue *prev;
  struct png_queue *next;
};

int insert_in_png_queue(struct png_queue *pngpacket);
void free_png_packet(struct png_queue *png);
struct png_queue* get_packet_from_queue(void);
void insert_in_evt_queue(struct evt_queue_elt *elt);

/* Parse an addr/port/ttl string. */
int hpt(char *h, u_int16 *port, unsigned char *ttl);

/* The library calls this function when it needs to schedule an event.
 *
 * Note that if we were running more than one stream at once we'd need
 * some way of corrolating the event to run with the context ID this
 * function gets passed.
 */
void RTPSchedule(context cid, rtp_opaque_t opaque, struct timeval *tp);

/* The library will call this callback (we tell it to, below) whenever some
   change occurs in a member's state. */
void update_member_callback(context cid, person id_no, rtpflag flags,
				   char *str);

/* The library will call this callback (we tell it to, below) whenever some
   change occurs in in the information about a member. */
void changed_memberinfo_callback(context cid, person id_no, memberinfo info, 
                                 char* str, rtpflag flags);

/* The library will call this callback (we tell it to, below) whenever two
 * members are observed to have the same CNAME (and thus are actually the same
 * member. */
void reverting_id_callback(context cid, person canonical, person to_delete,
                           void *userinfo_from_deletee, rtpflag flags);
/* The library will call this callback (we tell it to, below) whenever
 * a member's RTP or RTCP origination address changes */
void changed_member_address_callback(context cid, person id_no,
                                     char *addr, char *port, int is_rtcp);
/* The library will call this callback (we tell it to, below) whenever
 * there is an error in sending to the specified destination address.
 *
 * Note that errno is valid at the time this callback is called.
 *
 * BUG: when this happens, we ought to (depending on what errno is)
 * stop sending to this address. */
void send_error_callback(context cid, char *addr, u_int16 port, u_int8 ttl);

/* We'll call this when we receive any of SIGTERM, SIGHUP, or SIGQUIT.
 * We just set a flag, and let the main select loop bail us out, since
 * the RTP library is not re-entrant. */
void done(int sig);

/* Set up an RTP Library connection.
 * Return the maximum file descriptor number used, or -1 on error.
 * remote should contain the address to connect to in
 * host/port[/ttl] format.  (If ttl is not specified, '1' is used.)
 * cid will contain the context ID of the RTP session on sucessful return.
 * sock will contain the file descriptors of the RTP and RTCP ports.
 *
 * BUG: on failure, we should tear down the partially-completed
 * connections. */
int setup_connection(char *remote, context *cid, int sock[2]);

#endif
